@@ -1,4 +1,3 @@
-#!/usr/bin/python
 """
 Web application
 (from web.py)
@@ -13,8 +12,8 @@ import urllib
 import traceback
 import itertools
 import os
-import re
 import types
+from exceptions import SystemExit
 
 try:
     import wsgiref.handlers
@@ -43,10 +42,13 @@ class application:
     def __init__(self, mapping=(), fvars={}, autoreload=None):
         if autoreload is None:
             autoreload = web.config.get('debug', False)
-        self.mapping = mapping
+        self.init_mapping(mapping)
         self.fvars = fvars
         self.processors = []
-
+        
+        self.add_processor(loadhook(self._load))
+        self.add_processor(unloadhook(self._unload))
+        
         if autoreload:
             def main_module_name():
                 mod = sys.modules['__main__']
@@ -70,11 +72,11 @@ class application:
             
             def reload_mapping():
                 """loadhook to reload mapping and fvars."""
-                mod = __import__(module_name)
+                mod = __import__(module_name, None, None, [''])
                 mapping = getattr(mod, mapping_name, None)
                 if mapping:
                     self.fvars = mod.__dict__
-                    self.mapping = mapping
+                    self.init_mapping(mapping)
 
             self.add_processor(loadhook(Reloader()))
             if mapping_name and module_name:
@@ -86,10 +88,33 @@ class application:
                     __import__(main_module_name())
                 except ImportError:
                     pass
+                    
+    def _load(self):
+        web.ctx.app_stack.append(self)
+        
+    def _unload(self):
+        web.ctx.app_stack = web.ctx.app_stack[:-1]
+        
+        if web.ctx.app_stack:
+            # this is a sub-application, revert ctx to earlier state.
+            oldctx = web.ctx.get('_oldctx')
+            if oldctx:
+                web.ctx.home = oldctx.home
+                web.ctx.homepath = oldctx.homepath
+                web.ctx.path = oldctx.path
+                web.ctx.fullpath = oldctx.fullpath
+                
+    def _cleanup(self):
+        # Threads can be recycled by WSGI servers.
+        # Clearing up all thread-local state to avoid interefereing with subsequent requests.
+        utils.ThreadedDict.clear_all()
+
+    def init_mapping(self, mapping):
+        self.mapping = list(utils.group(mapping, 2))
 
     def add_mapping(self, pattern, classname):
-        self.mapping += (pattern, classname)
-        
+        self.mapping.append((pattern, classname))
+
     def add_processor(self, processor):
         """
         Adds a processor to the application. 
@@ -101,6 +126,7 @@ class application:
             ...
             >>>
             >>> def hello(handler): return "hello, " +  handler()
+            ...
             >>> app.add_processor(hello)
             >>> app.request("/web.py").data
             'hello, web.py'
@@ -175,7 +201,8 @@ class application:
         if 'HTTP_CONTENT_TYPE' in env:
             env['CONTENT_TYPE'] = env.pop('HTTP_CONTENT_TYPE')
 
-        if data:
+        if method in ["POST", "PUT"]:
+            data = data or ''
             import StringIO
             if isinstance(data, dict):
                 q = urllib.urlencode(data)
@@ -189,7 +216,7 @@ class application:
             response.status = status
             response.headers = dict(headers)
             response.header_items = headers
-        response.data = "".join(self.wsgifunc(cleanup_threadlocal=False)(env, start_response))
+        response.data = "".join(self.wsgifunc()(env, start_response))
         return response
 
     def browser(self):
@@ -203,7 +230,6 @@ class application:
     def handle_with_processors(self):
         def process(processors):
             try:
-                web.ctx.app_stack.append(self)
                 if processors:
                     p, processors = processors[0], processors[1:]
                     return p(lambda: process(processors))
@@ -211,17 +237,16 @@ class application:
                     return self.handle()
             except web.HTTPError:
                 raise
+            except (KeyboardInterrupt, SystemExit):
+                raise
             except:
                 print >> web.debug, traceback.format_exc()
                 raise self.internalerror()
-                    
-        try:
-            # processors must be applied in the resvere order. (??)
-            return process(self.processors)
-        finally:
-            web.ctx.app_stack = web.ctx.app_stack[:-1]
+        
+        # processors must be applied in the resvere order. (??)
+        return process(self.processors)
                         
-    def wsgifunc(self, *middleware, **kw):
+    def wsgifunc(self, *middleware):
         """Returns a WSGI-compatible function for this application."""
         def peep(iterator):
             """Peeps into an iterator by doing an iteration
@@ -234,42 +259,39 @@ class application:
                 firstchunk = iterator.next()
             except StopIteration:
                 firstchunk = ''
-            
+
             return itertools.chain([firstchunk], iterator)    
                                 
         def is_generator(x): return x and hasattr(x, 'next')
         
         def wsgi(env, start_resp):
-            self.load(env)
+            # clear threadlocal to avoid inteference of previous requests
+            self._cleanup()
 
+            self.load(env)
             try:
                 # allow uppercase methods only
                 if web.ctx.method.upper() != web.ctx.method:
                     raise web.nomethod()
 
                 result = self.handle_with_processors()
+                if is_generator(result):
+                    result = peep(result)
+                else:
+                    result = [result]
             except web.HTTPError, e:
-                result = e.data
+                result = [e.data]
 
-            if is_generator(result):
-                result = peep(result)
-            else:
-                result = [utils.utf8(result)]
+            result = web.safestr(iter(result))
 
             status, headers = web.ctx.status, web.ctx.headers
             start_resp(status, headers)
-
-            #@@@
-            # Since the CherryPy Webserver uses thread pool, the thread-local state is never cleared.
-            # This interferes with the other requests. 
-            # clearing the thread-local storage to avoid that.
-            # see utils.ThreadedDict for details
-            import threading
-            t = threading.currentThread()
-            if kw.get('cleanup_threadlocal', True) and hasattr(t, '_d'):
-                del t._d
-        
-            return result
+            
+            def cleanup():
+                self._cleanup()
+                yield '' # force this function to be a generator
+                            
+            return itertools.chain(result, cleanup())
 
         for m in middleware: 
             wsgi = m(wsgi)
@@ -332,6 +354,9 @@ class application:
         # http://trac.lighttpd.net/trac/ticket/406 requires:
         if env.get('SERVER_SOFTWARE', '').startswith('lighttpd/'):
             ctx.path = lstrips(env.get('REQUEST_URI').split('?')[0], ctx.homepath)
+            # Apache and CherryPy webservers unquote the url but lighttpd doesn't. 
+            # unquote explicitly for lighttpd to make ctx.path uniform across all servers.
+            ctx.path = urllib.unquote(ctx.path)
 
         if env.get('QUERY_STRING'):
             ctx.query = '?' + env.get('QUERY_STRING', '')
@@ -376,9 +401,8 @@ class application:
                         url += '?' + x
                 raise web.redirect(url)
             elif '.' in f:
-                x = f.split('.')
-                mod, cls = '.'.join(x[:-1]), x[-1]
-                mod = __import__(mod, globals(), locals(), [""])
+                mod, cls = f.rsplit('.', 1)
+                mod = __import__(mod, None, None, [''])
                 cls = getattr(mod, cls)
             else:
                 cls = fvars[f]
@@ -389,7 +413,7 @@ class application:
             return web.notfound()
 
     def _match(self, mapping, value):
-        for pat, what in utils.group(mapping, 2):
+        for pat, what in mapping:
             if isinstance(what, application):
                 if value.startswith(pat):
                     f = lambda: self._delegate_sub_application(pat, what)
@@ -402,7 +426,7 @@ class application:
                 result = utils.re_compile('^' + pat + '$').match(value)
                 
             if result: # it's a match
-                return what, [x and urllib.unquote(x) for x in result.groups()]
+                return what, [x for x in result.groups()]
         return None, None
         
     def _delegate_sub_application(self, dir, app):
@@ -412,18 +436,12 @@ class application:
         
         @@Any issues with when used with yield?
         """
-        try:
-            oldctx = web.storage(web.ctx)
-            web.ctx.home += dir
-            web.ctx.homepath += dir
-            web.ctx.path = web.ctx.path[len(dir):]
-            web.ctx.fullpath = web.ctx.fullpath[len(dir):]
-            return app.handle_with_processors()
-        finally:
-            web.ctx.home = oldctx.home
-            web.ctx.homepath = oldctx.homepath
-            web.ctx.path = oldctx.path
-            web.ctx.fullpath = oldctx.fullpath
+        web.ctx._oldctx = web.storage(web.ctx)
+        web.ctx.home += dir
+        web.ctx.homepath += dir
+        web.ctx.path = web.ctx.path[len(dir):]
+        web.ctx.fullpath = web.ctx.fullpath[len(dir):]
+        return app.handle_with_processors()
             
     def get_parent_app(self):
         if self in web.ctx.app_stack:
@@ -497,7 +515,7 @@ class subdomain_application(application):
         >>> class hello:
         ...     def GET(self): return "hello"
         >>>
-        >>> mapping = ("hello.example.com", app)
+        >>> mapping = (r"hello\.example\.com", app)
         >>> app2 = subdomain_application(mapping)
         >>> app2.request("/hello", host="hello.example.com").data
         'hello'
@@ -513,14 +531,14 @@ class subdomain_application(application):
         return self._delegate(fn, self.fvars, args)
         
     def _match(self, mapping, value):
-        for pat, what in utils.group(mapping, 2):
+        for pat, what in mapping:
             if isinstance(what, basestring):
                 what, result = utils.re_subm('^' + pat + '$', what, value)
             else:
                 result = utils.re_compile('^' + pat + '$').match(value)
 
             if result: # it's a match
-                return what, [x and urllib.unquote(x) for x in result.groups()]
+                return what, [x for x in result.groups()]
         return None, None
         
 def loadhook(h):
@@ -549,9 +567,31 @@ def unloadhook(h):
     """
     def processor(handler):
         try:
-            return handler()
-        finally:
+            result = handler()
+            is_generator = result and hasattr(result, 'next')
+        except:
+            # run the hook even when handler raises some exception
             h()
+            raise
+
+        if is_generator:
+            return wrap(result)
+        else:
+            h()
+            return result
+            
+    def wrap(result):
+        def next():
+            try:
+                return result.next()
+            except:
+                # call the hook at the and of iterator
+                h()
+                raise
+
+        result = iter(result)
+        while True:
+            yield next()
             
     return processor
 
@@ -586,28 +626,37 @@ def autodelegate(prefix=''):
             try:
                 return getattr(self, func)(*args)
             except TypeError:
-                return web.notfound()
+                raise web.notfound()
         else:
-            return web.notfound()
+            raise web.notfound()
     return internal
 
 class Reloader:
     """Checks to see if any loaded modules have changed on disk and, 
     if so, reloads them.
     """
+
+    SUFFIX = '$py.class' if sys.platform.startswith('java') else '.pyc'
+    """File suffix of compiled modules."""
+
     def __init__(self):
         self.mtimes = {}
 
     def __call__(self):
         for mod in sys.modules.values():
             self.check(mod)
-            
+
     def check(self, mod):
+        # jython registers java packages as modules but they either
+        # don't have a __file__ attribute or its value is None
+        if not (mod and hasattr(mod, '__file__') and mod.__file__):
+            return
+
         try: 
             mtime = os.stat(mod.__file__).st_mtime
-        except (AttributeError, OSError, IOError):
+        except (OSError, IOError):
             return
-        if mod.__file__.endswith('.pyc') and os.path.exists(mod.__file__[:-1]):
+        if mod.__file__.endswith(self.__class__.SUFFIX) and os.path.exists(mod.__file__[:-1]):
             mtime = max(os.stat(mod.__file__[:-1]).st_mtime, mtime)
             
         if mod not in self.mtimes:
@@ -618,7 +667,7 @@ class Reloader:
                 self.mtimes[mod] = mtime
             except ImportError: 
                 pass
-
+                
 if __name__ == "__main__":
     import doctest
     doctest.testmod()
