@@ -54,7 +54,7 @@ class UnknownParamstyle(Exception):
     """
     pass
     
-class SQLParam:
+class SQLParam(object):
     """
     Parameter in SQLQuery.
     
@@ -66,6 +66,8 @@ class SQLParam:
         >>> q.values()
         ['joe']
     """
+    __slots__ = ["value"]
+
     def __init__(self, value):
         self.value = value
         
@@ -95,7 +97,7 @@ class SQLParam:
 
 sqlparam =  SQLParam
 
-class SQLQuery:
+class SQLQuery(object):
     """
     You can pass this sort of thing as a clause in any db function.
     Otherwise, you can pass a dictionary to the keyword argument `vars`
@@ -104,8 +106,10 @@ class SQLQuery:
     Internally, consists of `items`, which is a list of strings and
     SQLParams, which get concatenated to produce the actual query.
     """
+    __slots__ = ["items"]
+
     # tested in sqlquote's docstring
-    def __init__(self, items=[]):
+    def __init__(self, items=None):
         r"""Creates a new SQLQuery.
         
             >>> SQLQuery("x")
@@ -118,7 +122,9 @@ class SQLQuery:
             >>> SQLQuery(SQLParam(1))
             <sql: '1'>
         """
-        if isinstance(items, list):
+        if items is None:
+            self.items = []
+        elif isinstance(items, list):
             self.items = items
         elif isinstance(items, SQLParam):
             self.items = [items]
@@ -132,6 +138,9 @@ class SQLQuery:
             if isinstance(item, SQLParam) and isinstance(item.value, SQLLiteral):
                 self.items[i] = item.value.v
 
+    def append(self, value):
+        self.items.append(value)
+
     def __add__(self, other):
         if isinstance(other, basestring):
             items = [other]
@@ -150,13 +159,12 @@ class SQLQuery:
         return SQLQuery(items + self.items)
 
     def __iadd__(self, other):
-        if isinstance(other, basestring):
-            items = [other]
+        if isinstance(other, (basestring, SQLParam)):
+            self.items.append(other)
         elif isinstance(other, SQLQuery):
-            items = other.items
+            self.items.extend(other.items)
         else:
             return NotImplemented
-        self.items.extend(items)
         return self
 
     def __len__(self):
@@ -195,21 +203,39 @@ class SQLQuery:
         """
         return [i.value for i in self.items if isinstance(i, SQLParam)]
         
-    def join(items, sep=' '):
+    def join(items, sep=' ', prefix=None, suffix=None, target=None):
         """
         Joins multiple queries.
         
         >>> SQLQuery.join(['a', 'b'], ', ')
         <sql: 'a, b'>
-        """
-        if len(items) == 0:
-            return SQLQuery("")
 
-        q = SQLQuery(items[0])
-        for item in items[1:]:
-            q += sep
-            q += item
-        return q
+        Optinally, prefix and suffix arguments can be provided.
+
+        >>> SQLQuery.join(['a', 'b'], ', ', prefix='(', suffix=')')
+        <sql: '(a, b)'>
+
+        If target argument is provided, the items are appended to target instead of creating a new SQLQuery.
+        """
+        if target is None:
+            target = SQLQuery()
+
+        target_items = target.items
+
+        if prefix:
+            target_items.append(prefix)
+
+        for i, item in enumerate(items):
+            if i != 0:
+                target_items.append(sep)
+            if isinstance(item, SQLQuery):
+                target_items.extend(item.items)
+            else:
+                target_items.append(item)
+
+        if suffix:
+            target_items.append(suffix)
+        return target
     
     join = staticmethod(join)
     
@@ -461,10 +487,9 @@ class DB:
         self.db_module = db_module
         self.keywords = keywords
 
-        
         self._ctx = threadeddict()
         # flag to enable/disable printing queries
-        self.printing = config.get('debug', False)
+        self.printing = config.get('debug_sql', config.get('debug', False))
         self.supports_multiple_insert = False
         
         try:
@@ -558,8 +583,8 @@ class DB:
         
         try:
             a = time.time()
-            paramstyle = getattr(self, 'paramstyle', 'pyformat')
-            out = cur.execute(sql_query.query(paramstyle), sql_query.values())
+            query, params = self._process_query(sql_query)
+            out = cur.execute(query, params)
             b = time.time()
         except:
             if self.printing:
@@ -573,6 +598,14 @@ class DB:
         if self.printing:
             print >> debug, '%s (%s): %s' % (round(b-a, 2), self.ctx.dbq_count, str(sql_query))
         return out
+
+    def _process_query(self, sql_query):
+        """Takes the SQLQuery object and returns query string and parameters.
+        """
+        paramstyle = getattr(self, 'paramstyle', 'pyformat')
+        query = sql_query.query(paramstyle)
+        params = sql_query.values()
+        return query, params
     
     def _where(self, where, vars): 
         if isinstance(where, (int, long)):
@@ -658,13 +691,21 @@ class DB:
             <sql: 'SELECT * FROM foo WHERE bar_id = 3'>
             >>> db.where('foo', source=2, crust='dewey', _test=True)
             <sql: "SELECT * FROM foo WHERE source = 2 AND crust = 'dewey'">
+            >>> db.where('foo', _test=True)
+            <sql: 'SELECT * FROM foo'>
         """
-        where = []
+        where_clauses = []
         for k, v in kwargs.iteritems():
-            where.append(k + ' = ' + sqlquote(v))
+            where_clauses.append(k + ' = ' + sqlquote(v))
+            
+        if where_clauses:
+            where = SQLQuery.join(where_clauses, " AND ")
+        else:
+            where = None
+            
         return self.select(table, what=what, order=order, 
                group=group, limit=limit, offset=offset, _test=_test, 
-               where=SQLQuery.join(where, ' AND '))
+               where=where)
     
     def sql_clauses(self, what, tables, where, group, order, limit, offset): 
         return (
@@ -718,7 +759,7 @@ class DB:
             _values = SQLQuery.join([sqlparam(v) for v in values.values()], ', ')
             sql_query = "INSERT INTO %s " % tablename + q(_keys) + ' VALUES ' + q(_values)
         else:
-            sql_query = SQLQuery("INSERT INTO %s DEFAULT VALUES" % tablename)
+            sql_query = SQLQuery(self._get_insert_default_values_query(tablename))
 
         if _test: return sql_query
         
@@ -744,6 +785,9 @@ class DB:
             self.ctx.commit()
         return out
         
+    def _get_insert_default_values_query(self, table):
+        return "INSERT INTO %s DEFAULT VALUES" % table
+
     def multiple_insert(self, tablename, values, seqname=None, _test=False):
         """
         Inserts multiple rows into `tablename`. The `values` must be a list of dictioanries, 
@@ -776,14 +820,13 @@ class DB:
             if v.keys() != keys:
                 raise ValueError, 'Bad data'
 
-        sql_query = SQLQuery('INSERT INTO %s (%s) VALUES ' % (tablename, ', '.join(keys))) 
-
-        data = []
-        for row in values:
-            d = SQLQuery.join([SQLParam(row[k]) for k in keys], ', ')
-            data.append('(' + d + ')')
-        sql_query += SQLQuery.join(data, ', ')
+        sql_query = SQLQuery('INSERT INTO %s (%s) VALUES ' % (tablename, ', '.join(keys)))
 
+        for i, row in enumerate(values):
+            if i != 0:
+                sql_query.append(", ")
+            SQLQuery.join([SQLParam(row[k]) for k in keys], sep=", ", target=sql_query, prefix="(", suffix=")")
+        
         if _test: return sql_query
 
         db_cursor = self._db_cursor()
@@ -855,8 +898,8 @@ class DB:
         where = self._where(where, vars)
 
         q = 'DELETE FROM ' + table
-        if where: q += ' WHERE ' + where
         if using: q += ' USING ' + sqllist(using)
+        if where: q += ' WHERE ' + where
 
         if _test: return q
 
@@ -915,7 +958,11 @@ class PostgresDB(DB):
 
     def _connect(self, keywords):
         conn = DB._connect(self, keywords)
-        conn.set_client_encoding('UTF8')
+        try:
+            conn.set_client_encoding('UTF8')
+        except AttributeError:
+            # fallback for pgdb driver
+            conn.cursor().execute("set client_encoding to 'UTF-8'")
         return conn
         
     def _connect_with_pooling(self, keywords):
@@ -942,6 +989,9 @@ class MySQLDB(DB):
         
     def _process_insert_query(self, query, tablename, seqname):
         return query, SQLQuery('SELECT last_insert_id();')
+        
+    def _get_insert_default_values_query(self, table):
+        return "INSERT INTO %s () VALUES()" % table
 
 def import_driver(drivers, preferred=None):
     """Import the first available driver or preferred driver.
@@ -962,6 +1012,10 @@ class SqliteDB(DB):
 
         if db.__name__ in ["sqlite3", "pysqlite2.dbapi2"]:
             db.paramstyle = 'qmark'
+            
+        # sqlite driver doesn't create datatime objects for timestamp columns unless `detect_types` option is passed.
+        # It seems to be supported in sqlite3 and pysqlite2 drivers, not surte about sqlite.
+        keywords.setdefault('detect_types', db.PARSE_DECLTYPES)
 
         self.paramstyle = db.paramstyle
         keywords['database'] = keywords.pop('db')
@@ -1019,6 +1073,16 @@ class MSSQLDB(DB):
         self.dbname = "mssql"
         DB.__init__(self, db, keywords)
 
+    def _process_query(self, sql_query):
+        """Takes the SQLQuery object and returns query string and parameters.
+        """
+        # MSSQLDB expects params to be a tuple. 
+        # Overwriting the default implementation to convert params to tuple.
+        paramstyle = getattr(self, 'paramstyle', 'pyformat')
+        query = sql_query.query(paramstyle)
+        params = sql_query.values()
+        return query, tuple(params)
+
     def sql_clauses(self, what, tables, where, group, order, limit, offset): 
         return (
             ('SELECT', what),
