@@ -1,6 +1,10 @@
 __all__ = ["runsimple"]
 
 import sys, os
+from SimpleHTTPServer import SimpleHTTPRequestHandler
+import urllib
+import posixpath
+
 import webapi as web
 import net
 import utils
@@ -134,92 +138,146 @@ def runsimple(func, server_address=("0.0.0.0", 8080)):
 
     [cp]: http://www.cherrypy.org
     """
-    from wsgiserver import CherryPyWSGIServer
-    from SimpleHTTPServer import SimpleHTTPRequestHandler
-    from BaseHTTPServer import BaseHTTPRequestHandler
-
-    class StaticApp(SimpleHTTPRequestHandler):
-        """WSGI application for serving static files."""
-        def __init__(self, environ, start_response):
-            self.headers = []
-            self.environ = environ
-            self.start_response = start_response
-
-        def send_response(self, status, msg=""):
-            self.status = str(status) + " " + msg
-
-        def send_header(self, name, value):
-            self.headers.append((name, value))
-
-        def end_headers(self):
-            pass
-
-        def log_message(*a): pass
-
-        def __iter__(self):
-            environ = self.environ
-
-            self.path = environ.get('PATH_INFO', '')
-            self.client_address = environ.get('REMOTE_ADDR','-'), \
-                                  environ.get('REMOTE_PORT','-')
-            self.command = environ.get('REQUEST_METHOD', '-')
-
-            from cStringIO import StringIO
-            self.wfile = StringIO() # for capturing error
-
-            f = self.send_head()
-            self.start_response(self.status, self.headers)
-
-            if f:
-                block_size = 16 * 1024
-                while True:
-                    buf = f.read(block_size)
-                    if not buf:
-                        break
-                    yield buf
-                f.close()
-            else:
-                value = self.wfile.getvalue()
-                yield value
-                    
-    class WSGIWrapper(BaseHTTPRequestHandler):
-        """WSGI wrapper for logging the status and serving static files."""
-        def __init__(self, app):
-            self.app = app
-            self.format = '%s - - [%s] "%s %s %s" - %s'
-
-        def __call__(self, environ, start_response):
-            def xstart_response(status, response_headers, *args):
-                write = start_response(status, response_headers, *args)
-                self.log(status, environ)
-                return write
-
-            path = environ.get('PATH_INFO', '')
-            if path.startswith('/static/'):
-                return StaticApp(environ, xstart_response)
-            else:
-                return self.app(environ, xstart_response)
-
-        def log(self, status, environ):
-            outfile = environ.get('wsgi.errors', web.debug)
-            req = environ.get('PATH_INFO', '_')
-            protocol = environ.get('ACTUAL_SERVER_PROTOCOL', '-')
-            method = environ.get('REQUEST_METHOD', '-')
-            host = "%s:%s" % (environ.get('REMOTE_ADDR','-'), 
-                              environ.get('REMOTE_PORT','-'))
-
-            #@@ It is really bad to extend from 
-            #@@ BaseHTTPRequestHandler just for this method
-            time = self.log_date_time_string()
-
-            msg = self.format % (host, time, protocol, method, req, status)
-            print >> outfile, utils.safestr(msg)
-            
-    func = WSGIWrapper(func)
-    server = CherryPyWSGIServer(server_address, func, server_name="localhost")
+    func = StaticMiddleware(func)
+    func = LogMiddleware(func)
+    
+    server = WSGIServer(server_address, func)
 
     print "http://%s:%d/" % server_address
     try:
         server.start()
     except KeyboardInterrupt:
         server.stop()
+
+def WSGIServer(server_address, wsgi_app):
+    """Creates CherryPy WSGI server listening at `server_address` to serve `wsgi_app`.
+    This function can be overwritten to customize the webserver or use a different webserver.
+    """
+    import wsgiserver
+    
+    # Default values of wsgiserver.ssl_adapters uses cheerypy.wsgiserver
+    # prefix. Overwriting it make it work with web.wsgiserver.
+    wsgiserver.ssl_adapters = {
+        'builtin': 'web.wsgiserver.ssl_builtin.BuiltinSSLAdapter',
+        'pyopenssl': 'web.wsgiserver.ssl_pyopenssl.pyOpenSSLAdapter',
+    }
+    
+    server = wsgiserver.CherryPyWSGIServer(server_address, wsgi_app, server_name="localhost")
+    server.nodelay = not sys.platform.startswith('java') # TCP_NODELAY isn't supported on the JVM
+    return server
+
+class StaticApp(SimpleHTTPRequestHandler):
+    """WSGI application for serving static files."""
+    def __init__(self, environ, start_response):
+        self.headers = []
+        self.environ = environ
+        self.start_response = start_response
+
+    def send_response(self, status, msg=""):
+        self.status = str(status) + " " + msg
+
+    def send_header(self, name, value):
+        self.headers.append((name, value))
+
+    def end_headers(self):
+        pass
+
+    def log_message(*a): pass
+
+    def __iter__(self):
+        environ = self.environ
+
+        self.path = environ.get('PATH_INFO', '')
+        self.client_address = environ.get('REMOTE_ADDR','-'), \
+                              environ.get('REMOTE_PORT','-')
+        self.command = environ.get('REQUEST_METHOD', '-')
+
+        from cStringIO import StringIO
+        self.wfile = StringIO() # for capturing error
+
+        try:
+            path = self.translate_path(self.path)
+            etag = '"%s"' % os.path.getmtime(path)
+            client_etag = environ.get('HTTP_IF_NONE_MATCH')
+            self.send_header('ETag', etag)
+            if etag == client_etag:
+                self.send_response(304, "Not Modified")
+                self.start_response(self.status, self.headers)
+                raise StopIteration
+        except OSError:
+            pass # Probably a 404
+
+        f = self.send_head()
+        self.start_response(self.status, self.headers)
+
+        if f:
+            block_size = 16 * 1024
+            while True:
+                buf = f.read(block_size)
+                if not buf:
+                    break
+                yield buf
+            f.close()
+        else:
+            value = self.wfile.getvalue()
+            yield value
+
+class StaticMiddleware:
+    """WSGI middleware for serving static files."""
+    def __init__(self, app, prefix='/static/'):
+        self.app = app
+        self.prefix = prefix
+        
+    def __call__(self, environ, start_response):
+        path = environ.get('PATH_INFO', '')
+        path = self.normpath(path)
+
+        if path.startswith(self.prefix):
+            return StaticApp(environ, start_response)
+        else:
+            return self.app(environ, start_response)
+
+    def normpath(self, path):
+        path2 = posixpath.normpath(urllib.unquote(path))
+        if path.endswith("/"):
+            path2 += "/"
+        return path2
+
+    
+class LogMiddleware:
+    """WSGI middleware for logging the status."""
+    def __init__(self, app):
+        self.app = app
+        self.format = '%s - - [%s] "%s %s %s" - %s'
+    
+        from BaseHTTPServer import BaseHTTPRequestHandler
+        import StringIO
+        f = StringIO.StringIO()
+        
+        class FakeSocket:
+            def makefile(self, *a):
+                return f
+        
+        # take log_date_time_string method from BaseHTTPRequestHandler
+        self.log_date_time_string = BaseHTTPRequestHandler(FakeSocket(), None, None).log_date_time_string
+        
+    def __call__(self, environ, start_response):
+        def xstart_response(status, response_headers, *args):
+            out = start_response(status, response_headers, *args)
+            self.log(status, environ)
+            return out
+
+        return self.app(environ, xstart_response)
+             
+    def log(self, status, environ):
+        outfile = environ.get('wsgi.errors', web.debug)
+        req = environ.get('PATH_INFO', '_')
+        protocol = environ.get('ACTUAL_SERVER_PROTOCOL', '-')
+        method = environ.get('REQUEST_METHOD', '-')
+        host = "%s:%s" % (environ.get('REMOTE_ADDR','-'), 
+                          environ.get('REMOTE_PORT','-'))
+
+        time = self.log_date_time_string()
+
+        msg = self.format % (host, time, protocol, method, req, status)
+        print >> outfile, utils.safestr(msg)
